use std::{
    borrow::Cow,
    fs::File,
    path::PathBuf,
    thread,
    time::{Duration, Instant},
};

use arg::Args;
use bitvec::{order::Msb0, view::AsBits};
use clap::Parser;
use colored::Colorize;
use console::Term;
use gif::{Encoder, Frame, Repeat};
use rapidriter_cat::next_frame;

const WIDTH: usize = 96;
const HEIGHT: usize = 38;

mod arg;

fn main() -> anyhow::Result<()> {
    let args = Args::parse();

    if let Some(path) = args.out {
        generate_gif(path)?;
    } else {
        // clear screen
        print!("\x1B[2J");

        if args.interactive {
            interactive_mode();
        } else {
            non_interactive_mode();
        }
    }

    Ok(())
}

fn generate_gif(path: PathBuf) -> anyhow::Result<()> {
    let mut gif = File::create(path)?;
    let mut encoder = Encoder::new(
        &mut gif,
        WIDTH.try_into().unwrap(),
        HEIGHT.try_into().unwrap(),
        &[0, 0, 0, 0xFF, 0, 0],
    )?;
    encoder.set_repeat(Repeat::Infinite)?;

    let mut frame = [0; 456];
    for i in 0..100 {
        let done = unsafe { next_frame(i, frame.as_mut_ptr()) } == 1;
        let mut exploded_frame = vec![];
        for b in frame.as_bits::<Msb0>() {
            exploded_frame.push(if *b { 1 } else { 0 });
        }

        let f = Frame {
            delay: 10, // 100 ms, or 10 FPS
            width: WIDTH.try_into().unwrap(),
            height: HEIGHT.try_into().unwrap(),
            buffer: Cow::Borrowed(&*exploded_frame),

            ..Default::default()
        };
        encoder.write_frame(&f)?;

        if done {
            break;
        }
    }

    Ok(())
}

fn interactive_mode() {
    let mut frame = [0; 456];

    let mut i = 0;

    let term = Term::stdout();

    loop {
        let has_next_frame = i < 99 && unsafe { next_frame(i, frame.as_mut_ptr()) } == 0;

        draw_frame(&frame);
        print!("Frame: {i}                  ");
        if !has_next_frame {
            print!(" (Last frame)");
        }
        println!("\np for previous frame, n for next frame, q to quit");

        loop {
            match term.read_char().expect("stdin error") {
                'n' if has_next_frame => {
                    i += 1;
                    break;
                }
                'p' if i > 0 => {
                    i -= 1;
                    break;
                }
                'q' => return,
                _ => {}
            }
        }
    }
}

fn non_interactive_mode() {
    let mut frame = [0; 456];

    let start = Instant::now();
    for i in 0..100u16 {
        let target = start + Duration::from_millis(u64::from(i) * 100);
        let now = Instant::now();
        if now < target {
            thread::sleep(target - now);
        }

        let done = unsafe { next_frame(i, frame.as_mut_ptr()) } == 1;

        draw_frame(&frame);
        println!("Frame: {}", i);

        if done {
            break;
        }
    }
}

fn draw_frame(frame: &[u8; 456]) {
    // reset cursor position to (1,1)
    print!("\x1B[1;1H");

    let bv = frame.as_bits::<Msb0>();

    // top part of border
    for _ in 0..(WIDTH + 2) {
        print!("{}", " ".on_white());
    }
    println!();

    for y in 0..HEIGHT {
        print!("{}", " ".on_white());

        for x in 0..WIDTH {
            if bv[x + y * WIDTH] {
                print!("{}", " ".on_red());
            } else {
                print!(" ");
            }
        }

        print!("{}", " ".on_white());

        println!();
    }

    // bottom part of border
    for _ in 0..(WIDTH + 2) {
        print!("{}", " ".on_white());
    }

    println!();
}
